# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

import sys
import os

from PyQt5 import uic
from PyQt5.QtWidgets import QSplitter
from PyQt5.QtWidgets import QStyle
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtGui import QIcon
# from PyQt5.QtCore import SIGNAL
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QDir
from PyQt5.QtCore import pyqtSignal

from ninja_ide import resources
from ninja_ide.core import file_manager
from ninja_ide.core import settings
from ninja_ide.core.filesystem_notifications import NinjaFileSystemWatcher
from ninja_ide.gui.main_panel import tab_widget
from ninja_ide.gui.editor import editor
from ninja_ide.gui.editor import highlighter
from ninja_ide.gui.editor import helpers
# from ninja_ide.gui.main_panel import browser_widget
from ninja_ide.gui.main_panel import start_page
from ninja_ide.gui.main_panel import image_viewer
from ninja_ide.tools import runner

from ninja_ide.tools.logger import NinjaLogger

logger = NinjaLogger('ninja_ide.gui.main_panel.main_container')

__mainContainerInstance = None


def MainContainer(*args, **kw):
    global __mainContainerInstance
    if __mainContainerInstance is None:
        __mainContainerInstance = __MainContainer(*args, **kw)
    return __mainContainerInstance


class __MainContainer(QSplitter):

###############################################################################
# MainContainer SIGNALS
###############################################################################
    """
    beforeFileSaved(QString)
    fileSaved(QString)
    currentTabChanged(QString)
    locateFunction(QString, QString, bool) [functionName, filePath, isVariable]
    openProject(QString)
    openPreferences()
    dontOpenStartPage()
    navigateCode(bool, int)
    addBackItemNavigation()
    updateLocator(QString)
    updateFileMetadata()
    findOcurrences(QString)
    cursorPositionChange(int, int)    #row, col
    fileOpened(QString)
    newFileOpened(QString)
    enabledFollowMode(bool)
    recentTabsModified(QStringList)
    migrationAnalyzed()
    allTabClosed()
    """
###############################################################################
    recentTabsModified = pyqtSignal(list)
    currentTabChanged = pyqtSignal(str)
    locateFunction = pyqtSignal(str, str, bool)
    navigateCode = pyqtSignal(bool, int)
    addBackItemNavigation = pyqtSignal()
    updateFileMetadata = pyqtSignal()
    updateLocator = pyqtSignal(str)
    openPreferences = pyqtSignal()
    dontOpenStartPage = pyqtSignal()
    allTabsClosed = pyqtSignal()
    findOcurrences = pyqtSignal(str)
    cursorPositionChanged = pyqtSignal(int, int)
    enabledFollowMode = pyqtSignal(bool)
    openProject = pyqtSignal(str)
    editorKeyPressEvent = pyqtSignal(object)
    beforeFileSaved = pyqtSignal(str)
    fileSaved = pyqtSignal(str)
    fileOpened = pyqtSignal(str)
    migrationAnalyzed = pyqtSignal()

    def __init__(self, parent=None):
        QSplitter.__init__(self, parent)
        self._parent = parent
        self._tabMain = tab_widget.TabWidget(self)
        self._tabSecondary = tab_widget.TabWidget(self)
        self.setAcceptDrops(True)
        self.addWidget(self._tabMain)
        self.addWidget(self._tabSecondary)
        self.setSizes([1, 1])
        self._tabSecondary.hide()
        self.actualTab = self._tabMain
        self._followMode = False
        self.splitted = False
        highlighter.restyle(resources.CUSTOM_SCHEME)
        #documentation browser
        self.docPage = None
        # File Watcher
        self._file_watcher = NinjaFileSystemWatcher
        self._watched_simple_files = []

        self._tabMain.currentChanged.connect(self._current_tab_changed)
        # self.connect(self._tabMain, SIGNAL("currentChanged(int)"),
        #     self._current_tab_changed)
        self._tabSecondary.currentChanged.connect(self._current_tab_changed)
        # self.connect(self._tabSecondary, SIGNAL("currentChanged(int)"),
        #     self._current_tab_changed)
        self._tabMain.currentChanged.connect(self._exit_follow_mode)
        # self.connect(self._tabMain, SIGNAL("currentChanged(int)"),
        #     self._exit_follow_mode)
        self._tabMain.changeActualTab.connect(self._change_actual)
        # self.connect(self._tabMain, SIGNAL("changeActualTab(QTabWidget)"),
        #     self._change_actual)
        self._tabSecondary.changeActualTab.connect(self._change_actual)
        # self.connect(self._tabSecondary, SIGNAL("changeActualTab(QTabWidget)"),
        #     self._change_actual)
        self._tabMain.splitTab.connect(self._split_this_tab)
        # self.connect(self._tabMain, SIGNAL("splitTab(QTabWidget, int, bool)"),
        #     self._split_this_tab)
        self._tabSecondary.splitTab.connect(self._split_this_tab)
        # self.connect(self._tabSecondary,
        #     SIGNAL("splitTab(QTabWidget, int, bool)"),
        #     self._split_this_tab)
        self._tabMain.reopenTab.connect(self._reopen_last_tab)
        # self.connect(self._tabMain, SIGNAL("reopenTab(QTabWidget, QString)"),
        #     self._reopen_last_tab)
        self._tabSecondary.reopenTab.connect(self._reopen_last_tab)
        # self.connect(self._tabSecondary,
        #     SIGNAL("reopenTab(QTabWidget, QString)"),
        #     self._reopen_last_tab)
        self._tabMain.syntaxChanged.connect(self._specify_syntax)
        # self.connect(self._tabMain, SIGNAL("syntaxChanged(QWidget, QString)"),
        #     self._specify_syntax)
        self._tabSecondary.syntaxChanged.connect(self._specify_syntax)
        # self.connect(self._tabSecondary,
        #     SIGNAL("syntaxChanged(QWidget, QString)"),
        #     self._specify_syntax)
        self._tabMain.allTabsClosed.connect(self._main_without_tabs)
        # self.connect(self._tabMain, SIGNAL("allTabsClosed()"),
        #     self._main_without_tabs)
        self._tabSecondary.allTabsClosed.connect(self._secondary_without_tabs)
        # self.connect(self._tabSecondary, SIGNAL("allTabsClosed()"),
        #     self._secondary_without_tabs)
        #reload file
        self._tabMain.reloadFile.connect(self.reload_file)
        # self.connect(self._tabMain, SIGNAL("reloadFile(QWidget)"),
        #     self.reload_file)
        self._tabSecondary.reloadFile.connect(self.reload_file)
        # self.connect(self._tabSecondary, SIGNAL("reloadFile(QWidget)"),
        #     self.reload_file)
        #for Save on Close operation
        self._tabMain.saveActualEditor.connect(self.save_file)
        # self.connect(self._tabMain, SIGNAL("saveActualEditor()"),
        #     self.save_file)
        self._tabSecondary.saveActualEditor.connect(self.save_file)
        # self.connect(self._tabSecondary, SIGNAL("saveActualEditor()"),
        #     self.save_file)
        #Navigate Code
        self._tabMain.navigateCode.connect(self._navigate_code)
        # self.connect(self._tabMain, SIGNAL("navigateCode(bool, int)"),
        #     self._navigate_code)
        self._tabSecondary.navigateCode.connect(self._navigate_code)
        # self.connect(self._tabSecondary, SIGNAL("navigateCode(bool, int)"),
        #     self._navigate_code)
        # Refresh recent tabs
        self._tabMain.recentTabsModified.connect(self._recent_files_changed)
        # self.connect(self._tabMain, SIGNAL("recentTabsModified(QStringList)"),
        #     self._recent_files_changed)

    def _recent_files_changed(self, files):
        self.recentTabsModified.emit(files)
        # self.emit(SIGNAL("recentTabsModified(QStringList)"), files)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        file_path = event.mimeData().urls()[0].toLocalFile()
        self.open_file(file_path)

    def _navigate_code(self, val, op):
        self.navigateCode.emit(val, op)
        # self.emit(SIGNAL("navigateCode(bool, int)"), val, op)

    def _main_without_tabs(self):
        if self._followMode:
            # if we were in follow mode, close the duplicated editor.
            self._exit_follow_mode()
        elif self._tabSecondary.isVisible():
            self.show_split(self.orientation())
        self.allTabsClosed.emit()
        # self.emit(SIGNAL("allTabsClosed()"))

    def _secondary_without_tabs(self):
        self.show_split(self.orientation())

    def _reopen_last_tab(self, tab, path):
        self.actualTab = tab
        self.open_file(path)

    def _change_actual(self, tabWidget):
        if not self._followMode:
            self.actualTab = tabWidget

    def _current_tab_changed(self, index):
        if self.actualTab.widget(index):
            self.currentTabChanged.emit(self.actualTab.widget(index)._id)
            # self.emit(SIGNAL("currentTabChanged(QString)"),
            #     self.actualTab.widget(index)._id)

    def split_tab(self, orientationHorizontal):
        """Split the main container in 2 areas.

        We are inverting the horizontal and vertical property here,
        because Qt see it Horizontal as side by side, but is confusing
        for the user."""
        if orientationHorizontal:
            self.show_split(Qt.Vertical)
        else:
            self.show_split(Qt.Horizontal)

    def _split_this_tab(self, tab, index, orientationHorizontal):
        tab.setCurrentIndex(index)
        if orientationHorizontal:
            self.show_split(Qt.Horizontal)
        else:
            self.show_split(Qt.Vertical)

    def change_tabs_visibility(self):
        if self._tabMain.tabBar().isVisible():
            self._tabMain.tabBar().hide()
            self._tabSecondary.tabBar().hide()
        else:
            self._tabMain.tabBar().show()
            self._tabSecondary.tabBar().show()

    def show_split(self, orientation):
        closingFollowMode = self._followMode
        if self._followMode:
            self._exit_follow_mode()
        if self._tabSecondary.isVisible() and \
        orientation == self.orientation():
            self._tabSecondary.hide()
            self.splitted = False
            for i in range(self._tabSecondary.count()):
                widget = self._tabSecondary.widget(0)
                name = self._tabSecondary.tabText(0)
                self._tabMain.add_tab(widget, name)
                if name in self._tabSecondary.titles:
                    self._tabSecondary.titles.remove(name)
                if type(widget) is editor.Editor and widget.textModified:
                    self._tabMain.tab_was_modified(True)
            self.actualTab = self._tabMain
        elif not self._tabSecondary.isVisible() and not closingFollowMode:
            widget = self.get_actual_widget()
            name = self._tabMain.tabText(self._tabMain.currentIndex())
            self._tabSecondary.add_tab(widget, name)
            if name in self._tabMain.titles:
                self._tabMain.titles.remove(name)
            if type(widget) is editor.Editor and widget.textModified:
                self._tabSecondary.tab_was_modified(True)
            self._tabSecondary.show()
            self.splitted = True
            self.setSizes([1, 1])
            self.actualTab = self._tabSecondary
            self.currentTabChanged.emit(widget.ID)
            # self.emit(SIGNAL("currentTabChanged(QString)"), widget.ID)
        self.setOrientation(orientation)

    def move_tab_to_next_split(self, tab_from):
        if self._followMode:
            return

        if tab_from == self._tabSecondary:
            tab_to = self._tabMain
        else:
            tab_to = self._tabSecondary

        widget = tab_from.currentWidget()
        name = tab_from.tabText(tab_from.currentIndex())
        tab_from.remove_title(tab_from.currentIndex())
        tab_to.add_tab(widget, name)
        if widget is editor.Editor and widget.textModified:
            tab_to.tab_was_saved(widget)
        tab_from.update_current_widget()

    def add_editor(self, fileName="", project=None, tabIndex=None,
                   syntax=None, use_open_highlight=False):
        project_obj = self._parent.explorer.get_project_given_filename(fileName)
        editorWidget = editor.create_editor(fileName=fileName, project=project,
            syntax=syntax, use_open_highlight=use_open_highlight,
            project_obj=project_obj)

        if not fileName:
            tabName = "New Document"
        else:
            tabName = file_manager.get_basename(fileName)

        #add the tab
        inserted_index = self.add_tab(editorWidget, tabName, tabIndex=tabIndex)
        self.actualTab.setTabToolTip(inserted_index,
            QDir.toNativeSeparators(fileName))
        #Connect signals
        editorWidget.modificationChanged.connect(self._editor_tab_was_modified)
        # self.connect(editorWidget, SIGNAL("modificationChanged(bool)"),
        #     self._editor_tab_was_modified)
        editorWidget.fileSaved.connect(self._editor_tab_was_saved)
        # self.connect(editorWidget, SIGNAL("fileSaved(QPlainTextEdit)"),
        #     self._editor_tab_was_saved)
        editorWidget.openDropFile.connect(self.open_file)
        # self.connect(editorWidget, SIGNAL("openDropFile(QString)"),
        #     self.open_file)
        editorWidget.addBackItemNavigation.connect(self.addBackItemNavigation.emit)
        # self.connect(editorWidget, SIGNAL("addBackItemNavigation()"),
        #     lambda: self.emit(SIGNAL("addBackItemNavigation()")))
        editorWidget.locateFunction.connect(self._editor_locate_function)
        # self.connect(editorWidget,
        #     SIGNAL("locateFunction(QString, QString, bool)"),
        #     self._editor_locate_function)
        editorWidget.warningsFound.connect(self._show_warning_tab_indicator)
        # self.connect(editorWidget, SIGNAL("warningsFound(QPlainTextEdit)"),
        #     self._show_warning_tab_indicator)
        editorWidget.errorsFound.connect(self._show_error_tab_indicator)
        # self.connect(editorWidget, SIGNAL("errorsFound(QPlainTextEdit)"),
        #     self._show_error_tab_indicator)
        editorWidget.cleanDocument.connect(self._hide_icon_tab_indicator)
        # self.connect(editorWidget, SIGNAL("cleanDocument(QPlainTextEdit)"),
        #     self._hide_icon_tab_indicator)
        editorWidget.findOcurrences.connect(self._find_occurrences)
        # self.connect(editorWidget, SIGNAL("findOcurrences(QString)"),
        #     self._find_occurrences)
        editorWidget.migrationAnalyzed.connect(self.migrationAnalyzed.emit)
        # self.connect(editorWidget, SIGNAL("migrationAnalyzed()"),
        #     lambda: self.emit(SIGNAL("migrationAnalyzed()")))
        #Cursor position changed
        editorWidget.cursorPositionChange.connect(self._cursor_position_changed)
        # self.connect(editorWidget, SIGNAL("cursorPositionChange(int, int)"),
        #     self._cursor_position_changed)
        #keyPressEventSignal for plugins
        editorWidget.keyPressed.connect(self._editor_keyPressEvent)
        # self.connect(editorWidget, SIGNAL("keyPressEvent(QEvent)"),
        #     self._editor_keyPressEvent)

        #emit a signal about the file open
        self.fileOpened.emit(fileName)
        # self.emit(SIGNAL("fileOpened(QString)"), fileName)

        return editorWidget

    def update_editor_project(self):
        for i in range(self._tabMain.count()):
            widget = self._tabMain.widget(i)
            if type(widget) is editor.Editor:
                project = self._parent.explorer.get_project_given_filename(
                    widget.ID)
                widget.set_project(project)
        for i in range(self._tabSecondary.count()):
            widget = self._tabSecondary.widget(i)
            if type(widget) is editor.Editor:
                project = self._parent.explorer.get_project_given_filename(
                    widget.ID)
                widget.set_project(project)

    def reset_pep8_warnings(self, value):
        for i in range(self._tabMain.count()):
            widget = self._tabMain.widget(i)
            if type(widget) is editor.Editor:
                if value:
                    widget.syncDocErrorsSignal = True
                    widget.pep8.check_style()
                else:
                    widget.hide_pep8_errors()
        for i in range(self._tabSecondary.count()):
            widget = self._tabSecondary.widget(i)
            if type(widget) is editor.Editor:
                if value:
                    widget.syncDocErrorsSignal = True
                    widget.pep8.check_style()
                else:
                    widget.hide_pep8_errors()

    def reset_lint_warnings(self, value):
        for i in range(self._tabMain.count()):
            widget = self._tabMain.widget(i)
            if type(widget) is editor.Editor:
                if value:
                    widget.syncDocErrorsSignal = True
                    widget.errors.check_errors()
                else:
                    widget.hide_lint_errors()
        for i in range(self._tabSecondary.count()):
            widget = self._tabSecondary.widget(i)
            if type(widget) is editor.Editor:
                if value:
                    widget.syncDocErrorsSignal = True
                    widget.errors.check_errors()
                else:
                    widget.hide_lint_errors()

    def _cursor_position_changed(self, row, col):
        # self.emit(SIGNAL("cursorPositionChange(int, int)"), row, col)
        self.cursorPositionChanged.emit(row, col)

    def _find_occurrences(self, word):
        # self.emit(SIGNAL("findOcurrences(QString)"), word)
        self.findOcurrences.emit(word)

    def _show_warning_tab_indicator(self, editorWidget):
        index = self.actualTab.indexOf(editorWidget)
        # self.emit(SIGNAL("updateFileMetadata()"))
        self.updateFileMetadata.emit()
        if index >= 0:
            self.actualTab.setTabIcon(index,
                QIcon(self.style().standardIcon(QStyle.SP_MessageBoxWarning)))

    def _show_error_tab_indicator(self, editorWidget):
        index = self.actualTab.indexOf(editorWidget)
        # self.emit(SIGNAL("updateFileMetadata()"))
        self.updateFileMetadata.emit()
        if index >= 0:
            self.actualTab.setTabIcon(index,
                QIcon(resources.IMAGES['bug']))

    def _hide_icon_tab_indicator(self, editorWidget):
        index = self.actualTab.indexOf(editorWidget)
        # self.emit(SIGNAL("updateFileMetadata()"))
        self.updateFileMetadata.emit()
        if index >= 0:
            self.actualTab.setTabIcon(index, QIcon())

    def _editor_keyPressEvent(self, event):
        # self.emit(SIGNAL("editorKeyPressEvent(QEvent)"), event)
        self.editorKeyPressEvent.emit(event)

    def _editor_locate_function(self, function, filePath, isVariable):
        # self.emit(SIGNAL("locateFunction(QString, QString, bool)"),
        #     function, filePath, isVariable)
        self.locateFunction.emit(function, filePath, isVariable)

    def _editor_tab_was_modified(self, val=True):
        self.actualTab.tab_was_modified(val)

    def _editor_tab_was_saved(self, editorWidget=None):
        self.actualTab.tab_was_saved(editorWidget)
        # self.emit(SIGNAL("updateLocator(QString)"), editorWidget.ID)
        self.updateLocator.emit(editorWidget.ID)

    def add_tab(self, widget, tabName, tabIndex=None):
        return self.actualTab.add_tab(widget, tabName, index=tabIndex)

    def get_actual_widget(self):
        return self.actualTab.currentWidget()

    def get_actual_editor(self):
        """Return the Actual Editor or None

        Return an instance of Editor if the Current Tab contains
        an Editor or None if it is not an instance of Editor"""
        widget = self.actualTab.currentWidget()
        if type(widget) is editor.Editor:
            return widget
        return None

    def reload_file(self, editorWidget=None):
        if editorWidget is None:
            editorWidget = self.get_actual_editor()
        if editorWidget is not None and editorWidget.ID:
            fileName = editorWidget.ID
            self._file_watcher.allow_kill = False
            old_cursor_position = editorWidget.textCursor().position()
            old_widget_index = self.actualTab.indexOf(editorWidget)
            self.actualTab.removeTab(old_widget_index)
            #open the file in the same tab as before
            self.open_file(fileName, tabIndex=old_widget_index)
            #get the new editor and set the old cursor position
            editorWidget = self.get_actual_editor()
            cursor = editorWidget.textCursor()
            cursor.setPosition(old_cursor_position)
            editorWidget.setTextCursor(cursor)
            self._file_watcher.allow_kill = True

    def open_image(self, fileName):
        try:
            if not self.is_open(fileName):
                viewer = image_viewer.ImageViewer(fileName)
                self.add_tab(viewer, file_manager.get_basename(fileName))
                viewer.id = fileName
            else:
                self.move_to_open(fileName)
        except Exception as reason:
            logger.error('open_image: %s', reason)
            QMessageBox.information(self, self.tr("Incorrect File"),
                self.tr("The image couldn\'t be open"))

    def open_file(self, filename='', cursorPosition=-1,
                    tabIndex=None, positionIsLineNumber=False, notStart=True):
        if not filename:
            if settings.WORKSPACE:
                directory = settings.WORKSPACE
            else:
                directory = os.path.expanduser("~")
                editorWidget = self.get_actual_editor()
                pexplorer = self._parent.explorer
                current_project = pexplorer and pexplorer.get_actual_project()
                if current_project is not None:
                    directory = current_project
                elif editorWidget is not None and editorWidget.ID:
                    directory = file_manager.get_folder(editorWidget.ID)
            extensions = ';;'.join(
                ['(*%s)' % e for e in
                    settings.SUPPORTED_EXTENSIONS + ['.*', '']])
            fileNames = QFileDialog.getOpenFileNames(self,
                self.tr("Open File"), directory, extensions)[0]
        else:
            fileNames = [filename]
        if not fileNames:
            return

        for filename in fileNames:
            if file_manager.get_file_extension(filename) in ('jpg', 'png'):
                self.open_image(filename)
            elif file_manager.get_file_extension(filename).endswith('ui'):
                self.w = uic.loadUi(filename)
                self.w.show()
            else:
                self.__open_file(filename, cursorPosition,
                    tabIndex, positionIsLineNumber, notStart)

    def __open_file(self, fileName='', cursorPosition=-1,
                    tabIndex=None, positionIsLineNumber=False, notStart=True):
        try:
            if not self.is_open(fileName):
                self.actualTab.notOpening = False
                content = file_manager.read_file_content(fileName)
                editorWidget = self.add_editor(fileName, tabIndex=tabIndex,
                    use_open_highlight=True)

                #Add content
                #we HAVE to add the editor's content before set the ID
                #because of the minimap logic
                editorWidget.setPlainText(content)
                editorWidget.ID = fileName
                editorWidget.async_highlight()
                encoding = file_manager.get_file_encoding(content)
                editorWidget.encoding = encoding
                if cursorPosition == -1:
                    cursorPosition = 0
                if not positionIsLineNumber:
                    editorWidget.set_cursor_position(cursorPosition)
                else:
                    editorWidget.go_to_line(cursorPosition)
                self.add_standalone_watcher(editorWidget.ID, notStart)

                #New file then try to add a coding line
                if not content:
                    helpers.insert_coding_line(editorWidget)
                    self.save_file(editorWidget=editorWidget)

                if not editorWidget.has_write_permission():
                    fileName += self.tr(" (Read-Only)")
                    index = self.actualTab.currentIndex()
                    self.actualTab.setTabText(index, fileName)

            else:
                self.move_to_open(fileName)
                editorWidget = self.get_actual_editor()
                if editorWidget and notStart and cursorPosition != -1:
                    if positionIsLineNumber:
                        editorWidget.go_to_line(cursorPosition)
                    else:
                        editorWidget.set_cursor_position(cursorPosition)
            # self.emit(SIGNAL("currentTabChanged(QString)"), fileName)
            self.currentTabChanged.emit(fileName)
        except file_manager.NinjaIOException as reason:
            if notStart:
                QMessageBox.information(self,
                    self.tr("The file couldn't be open"), str(reason))
        except Exception as reason:
            logger.error('open_file: %s', reason)
        self.actualTab.notOpening = True

    def add_standalone_watcher(self, filename, not_start=True):
        # Add File Watcher if needed
        opened_projects = self._parent.explorer.get_opened_projects()
        opened_projects = [p.path for p in opened_projects]
        #alone = not_start
        #for folder in opened_projects:
            #if file_manager.belongs_to_folder(folder, filename):
                #alone = False
        #if alone or sys.platform == 'darwin':
            #self._file_watcher.add_file_watch(filename)
            #self._watched_simple_files.append(filename)

    def remove_standalone_watcher(self, filename):
        if filename in self._watched_simple_files:
            self._file_watcher.remove_file_watch(filename)
            self._watched_simple_files.remove(filename)

    def is_open(self, filename):
        return self._tabMain.is_open(filename) != -1 or \
            self._tabSecondary.is_open(filename) != -1

    def move_to_open(self, filename):
        if self._tabMain.is_open(filename) != -1:
            self._tabMain.move_to_open(filename)
            self.actualTab = self._tabMain
        elif self._tabSecondary.is_open(filename) != -1:
            self._tabSecondary.move_to_open(filename)
            self.actualTab = self._tabSecondary
        self.actualTab.currentWidget().setFocus()
        # self.emit(SIGNAL("currentTabChanged(QString)"), filename)
        self.currentTabChanged.emit(filename)

    def get_widget_for_path(self, filename):
        if self._tabMain.is_open(filename) != -1:
            index = self._tabMain.search_for_identifier_index(filename)
            return self._tabMain.widget(index)
        elif self._tabSecondary.is_open(filename) != -1:
            index = self._tabSecondary.search_for_identifier_index(filename)
            return self._tabSecondary.widget(index)
        return None

    def change_open_tab_name(self, id, newId):
        """Search for the Tab with id, and set the newId to that Tab."""
        index = self._tabMain.is_open(id)
        if index != -1:
            widget = self._tabMain.widget(index)
            tabContainer = self._tabMain
        elif self._tabSecondary.is_open(id):
            # tabSecondaryIndex is recalculated because there is a really
            # small chance that the tab is there, so there is no need to
            # calculate this value by default
            index = self._tabSecondary.is_open(id)
            widget = self._tabSecondary.widget(index)
            tabContainer = self._tabSecondary
        tabName = file_manager.get_basename(newId)
        tabContainer.change_open_tab_name(index, tabName)
        widget.ID = newId

    def close_deleted_file(self, id):
        """Search for the Tab with id, and ask the user if should be closed."""
        index = self._tabMain.is_open(id)
        if index != -1:
            tabContainer = self._tabMain
        elif self._tabSecondary.is_open(id):
            # tabSecondaryIndex is recalculated because there is a really
            # small chance that the tab is there, so there is no need to
            # calculate this value by default
            index = self._tabSecondary.is_open(id)
            tabContainer = self._tabSecondary
        result = QMessageBox.question(self, self.tr("Close Deleted File"),
            self.tr("Are you sure you want to close the deleted file?\n"
                    "The content will be completely deleted."),
            buttons=QMessageBox.Yes | QMessageBox.No)
        if result == QMessageBox.Yes:
            tabContainer.removeTab(index)

    def save_file(self, editorWidget=None):
        if not editorWidget:
            editorWidget = self.get_actual_editor()
        if not editorWidget:
            return False
        try:
            editorWidget.just_saved = True
            if editorWidget.newDocument or \
            not file_manager.has_write_permission(editorWidget.ID):
                return self.save_file_as()

            fileName = editorWidget.ID
            # self.emit(SIGNAL("beforeFileSaved(QString)"), fileName)
            self.beforeFileSaved.emit(fileName)
            if settings.REMOVE_TRAILING_SPACES:
                helpers.remove_trailing_spaces(editorWidget)
            content = editorWidget.get_text()
            file_manager.store_file_content(
                fileName, content, addExtension=False)
            self._file_watcher.allow_kill = False
            if editorWidget.ID != fileName:
                self.remove_standalone_watcher(editorWidget.ID)
            self.add_standalone_watcher(fileName)
            self._file_watcher.allow_kill = True
            editorWidget.ID = fileName
            encoding = file_manager.get_file_encoding(content)
            editorWidget.encoding = encoding
            # self.emit(SIGNAL("fileSaved(QString)"),
            #     (self.tr("File Saved: %s") % fileName))
            self.fileSaved.emit(self.tr("File Saved: %s" % fileName))
            editorWidget._file_saved()
            return True
        except Exception as reason:
            editorWidget.just_saved = False
            logger.error('save_file: %s', reason)
            QMessageBox.information(self, self.tr("Save Error"),
                self.tr("The file couldn't be saved!"))
        return False

    def save_file_as(self):
        editorWidget = self.get_actual_editor()
        if not editorWidget:
            return False
        try:
            editorWidget.just_saved = True
            filters = '(*.py);;(*.*)'
            if editorWidget.ID:
                ext = file_manager.get_file_extension(editorWidget.ID)
                if ext != 'py':
                    filters = '(*.%s);;(*.py);;(*.*)' % ext
            save_folder = self._get_save_folder(editorWidget.ID)
            fileName = QFileDialog.getSaveFileName(
                self._parent, self.tr("Save File"), save_folder, filters)[0]
            if not fileName:
                return False

            if settings.REMOVE_TRAILING_SPACES:
                helpers.remove_trailing_spaces(editorWidget)
            newFile = file_manager.get_file_extension(fileName) == ''
            fileName = file_manager.store_file_content(
                fileName, editorWidget.get_text(),
                addExtension=True, newFile=newFile)
            self.actualTab.setTabText(self.actualTab.currentIndex(),
                file_manager.get_basename(fileName))
            editorWidget.register_syntax(
                file_manager.get_file_extension(fileName))
            self._file_watcher.allow_kill = False
            if editorWidget.ID != fileName:
                self.remove_standalone_watcher(editorWidget.ID)
            editorWidget.ID = fileName
            # self.emit(SIGNAL("fileSaved(QString)"),
            #     (self.tr("File Saved: %s") % fileName))
            self.fileSaved.emit(self.tr("File Saved: %s" % fileName))
            # self.emit(SIGNAL("currentTabChanged(QString)"), fileName)
            self.currentTabChanged.emit(fileName)
            editorWidget._file_saved()
            self.add_standalone_watcher(fileName)
            self._file_watcher.allow_kill = True
            return True
        except file_manager.NinjaFileExistsException as ex:
            editorWidget.just_saved = False
            QMessageBox.information(self, self.tr("File Already Exists"),
                (self.tr("Invalid Path: the file '%s' already exists.") %
                    ex.filename))
        except Exception as reason:
            editorWidget.just_saved = False
            logger.error('save_file_as: %s', reason)
            QMessageBox.information(self, self.tr("Save Error"),
                self.tr("The file couldn't be saved!"))
            self.actualTab.setTabText(self.actualTab.currentIndex(),
                self.tr("New Document"))
        return False

    def _get_save_folder(self, fileName):
        """
        Returns the root directory of the 'Main Project' or the home folder
        """
        actual_project = self._parent.explorer.get_actual_project()
        if actual_project:
            return actual_project
        return os.path.expanduser("~")

    def save_project(self, projectFolder):
        for i in range(self._tabMain.count()):
            editorWidget = self._tabMain.widget(i)
            if type(editorWidget) is editor.Editor and \
            file_manager.belongs_to_folder(projectFolder, editorWidget.ID):
                reloaded = self._tabMain.check_for_external_modifications(
                    editorWidget)
                if not reloaded:
                    self.save_file(editorWidget)
        for i in range(self._tabSecondary.count()):
            editorWidget = self._tabSecondary.widget(i)
            if type(editorWidget) is editor.Editor and \
            file_manager.belongs_to_folder(projectFolder, editorWidget.ID):
                reloaded = self._tabSecondary.check_for_external_modifications(
                    editorWidget)
                if not reloaded:
                    self.save_file(editorWidget)

    def save_all(self):
        for i in range(self._tabMain.count()):
            editorWidget = self._tabMain.widget(i)
            if type(editorWidget) is editor.Editor:
                reloaded = self._tabMain.check_for_external_modifications(
                    editorWidget)
                if not reloaded:
                    self.save_file(editorWidget)
        for i in range(self._tabSecondary.count()):
            editorWidget = self._tabSecondary.widget(i)
            self._tabSecondary.check_for_external_modifications(editorWidget)
            if type(editorWidget) is editor.Editor:
                reloaded = self._tabSecondary.check_for_external_modifications(
                    editorWidget)
                if not reloaded:
                    self.save_file(editorWidget)

    def call_editors_function(self, call_function, *arguments):
        args = arguments[0]
        kwargs = arguments[1]
        for i in range(self._tabMain.count()):
            editorWidget = self._tabMain.widget(i)
            if type(editorWidget) is editor.Editor:
                function = getattr(editorWidget, call_function)
                function(*args, **kwargs)
        for i in range(self._tabSecondary.count()):
            editorWidget = self._tabSecondary.widget(i)
            self._tabSecondary.check_for_external_modifications(editorWidget)
            if type(editorWidget) is editor.Editor:
                function = getattr(editorWidget, call_function)
                function(*args, **kwargs)

    def show_start_page(self):
        if not self.is_open("Start Page"):
            startPage = start_page.StartPage(parent=self)
            startPage.openProject.connect(self.open_project)
            # self.connect(startPage, SIGNAL("openProject(QString)"),
            #     self.open_project)
            startPage.openPreferences.connect(self.openPreferences.emit)
            # self.connect(startPage, SIGNAL("openPreferences()"),
            #     lambda: self.emit(SIGNAL("openPreferences()")))
            self.add_tab(startPage, 'Start Page')
        else:
            self.move_to_open("Start Page")

    def show_python_doc(self):
        # if sys.platform == 'win32':
        #     self.docPage = browser_widget.BrowserWidget(
        #         'http://docs.python.org/')
        #     self.add_tab(self.docPage, self.tr("Python Documentation"))
        # else:
        #     process = runner.start_pydoc()
        #     self.docPage = browser_widget.BrowserWidget(process[1], process[0])
        #     self.add_tab(self.docPage, self.tr("Python Documentation"))
        pass

    def editor_jump_to_line(self, lineno=None):
        """Jump to line *lineno* if it is not None
        otherwise ask to the user the line number to jump
        """
        editorWidget = self.get_actual_editor()
        if editorWidget:
            editorWidget.jump_to_line(lineno=lineno)

    def show_follow_mode(self):
        tempTab = self.actualTab
        self.actualTab = self._tabMain
        editorWidget = self.get_actual_editor()
        if not editorWidget:
            return
        if self._tabSecondary.isVisible() and not self._followMode:
            self.show_split(self.orientation())
        if self._followMode:
            self._exit_follow_mode()
        else:
            self._followMode = True
            self.setOrientation(Qt.Horizontal)
            name = self._tabMain.tabText(self._tabMain.currentIndex())
            editor2 = editor.create_editor()
            editor2.setDocument(editorWidget.document())
            self._tabSecondary.add_tab(editor2, name)
            if editorWidget.textModified:
                self._tabSecondary.tab_was_modified(True)
            self._tabSecondary.show()
            editor2.verticalScrollBar().setRange(
                editorWidget._sidebarWidget.highest_line - 2, 0)
            self._tabSecondary.setTabsClosable(False)
            self._tabSecondary.follow_mode = True
            self.setSizes([1, 1])
            # self.emit(SIGNAL("enabledFollowMode(bool)"), self._followMode)
            self.enabledFollowMode.emit(self._followMode)
        self.actualTab = tempTab

    def _exit_follow_mode(self):
        if self._followMode:
            self._followMode = False
            self._tabSecondary.close_tab()
            self._tabSecondary.hide()
            self._tabSecondary.follow_mode = False
            self._tabSecondary.setTabsClosable(True)
            # self.emit(SIGNAL("enabledFollowMode(bool)"), self._followMode)
            self.enabledFollowMode.emit(self._followMode)

    def get_opened_documents(self):
        if self._followMode:
            return [self._tabMain.get_documents_data(), []]
        return [self._tabMain.get_documents_data(),
            self._tabSecondary.get_documents_data()]

    def open_files(self, files, mainTab=True, notIDEStart=True):
        if mainTab:
            self.actualTab = self._tabMain
        else:
            self.actualTab = self._tabSecondary
            if files:
                self._tabSecondary.show()
                self.splitted = True

        for fileData in files:
            if file_manager.file_exists(fileData[0]):
                self.open_file(fileData[0], fileData[1], notStart=notIDEStart)
        self.actualTab = self._tabMain

    def check_for_unsaved_tabs(self):
        return self._tabMain._check_unsaved_tabs() or \
            self._tabSecondary._check_unsaved_tabs()

    def get_unsaved_files(self):
        return self._tabMain.get_unsaved_files() or \
            self._tabSecondary.get_unsaved_files()

    def reset_editor_flags(self):
        for i in range(self._tabMain.count()):
            widget = self._tabMain.widget(i)
            if type(widget) is editor.Editor:
                widget.set_flags()
        for i in range(self._tabSecondary.count()):
            widget = self._tabSecondary.widget(i)
            if type(widget) is editor.Editor:
                widget.set_flags()

    def _specify_syntax(self, widget, syntaxLang):
        if type(widget) is editor.Editor:
            widget.restyle(syntaxLang)

    def apply_editor_theme(self, family, size):
        for i in range(self._tabMain.count()):
            widget = self._tabMain.widget(i)
            if type(widget) is editor.Editor:
                widget.restyle()
                widget.set_font(family, size)
        for i in range(self._tabSecondary.count()):
            widget = self._tabSecondary.widget(i)
            if type(widget) is editor.Editor:
                widget.restyle()
                widget.set_font(family, size)

    def update_editor_margin_line(self):
        for i in range(self._tabMain.count()):
            widget = self._tabMain.widget(i)
            if type(widget) is editor.Editor:
                widget._update_margin_line()
        for i in range(self._tabSecondary.count()):
            widget = self._tabSecondary.widget(i)
            if type(widget) is editor.Editor:
                widget._update_margin_line()

    def open_project(self, path):
        # self.emit(SIGNAL("openProject(QString)"), path)
        self.openProject.emit(path)

    def close_python_doc(self):
        #close the python document server (if running)
        if self.docPage:
            index = self.actualTab.indexOf(self.docPage)
            self.actualTab.removeTab(index)
            #assign None to the browser
            self.docPage = None

    def close_tab(self):
        """Close the current tab in the current TabWidget."""
        self.actualTab.close_tab()

    def change_tab(self):
        """Change the tab in the current TabWidget."""
        self.actualTab.change_tab()

    def change_tab_reverse(self):
        """Change the tab in the current TabWidget backwards."""
        self.actualTab.change_tab_reverse()

    def show_navigation_buttons(self):
        self.actualTab.navigator.show_menu_navigation()

    def change_split_focus(self):
        if self.actualTab == self._tabMain and self._tabSecondary.isVisible():
            self.actualTab = self._tabSecondary
        else:
            self.actualTab = self._tabMain
        widget = self.actualTab.currentWidget()
        if widget is not None:
            widget.setFocus()



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

from PyQt5.QtWidgets import QWidget
# from PyQt5.QtCore import SIGNAL
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtWidgets import QSpacerItem
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QListWidget
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QHBoxLayout

from ninja_ide.core import file_manager
from ninja_ide.gui.main_panel import itab_item


class TabGroup(QWidget, itab_item.ITabItem):

    def __init__(self, project, name, actions):
        QWidget.__init__(self)
        itab_item.ITabItem.__init__(self)
        vbox = QVBoxLayout(self)
        self.actions = actions
        self.project = project
        self.ID = self.project
        self.name = name
        self.tabs = []
        self.listWidget = QListWidget()
        hbox = QHBoxLayout()
        btnExpand = QPushButton(self.tr("Expand this Files"))
        btnExpandAll = QPushButton(self.tr("Expand all Groups"))
        hbox.addWidget(btnExpandAll)
        hbox.addSpacerItem(QSpacerItem(20, 20, QSizePolicy.Expanding))
        hbox.addWidget(btnExpand)
        vbox.addLayout(hbox)
        vbox.addWidget(self.listWidget)

        self.connect(btnExpand, SIGNAL("clicked()"), self.expand_this)
        self.connect(btnExpandAll, SIGNAL("clicked()"),
            self.actions.deactivate_tabs_groups)

    def add_widget(self, widget):
        self.tabs.append(widget)
        self.listWidget.addItem(widget.ID)

    def expand_this(self):
        self.actions.group_tabs_together()
        for tab in self.tabs:
            tabName = file_manager.get_basename(tab.ID)
            self.actions.ide.mainContainer.add_tab(tab, tabName)
        index = self.actions.ide.mainContainer._tabMain.indexOf(self)
        self.actions.ide.mainContainer._tabMain.removeTab(index)
        self.tabs = []
        self.listWidget.clear()

    def only_expand(self):
        for tab in self.tabs:
            tabName = file_manager.get_basename(tab.ID)
            self.actions.ide.mainContainer.add_tab(tab, tabName)
        index = self.actions.ide.mainContainer._tabMain.indexOf(self)
        self.actions.ide.mainContainer._tabMain.removeTab(index)
        self.tabs = []
        self.listWidget.clear()



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import
from __future__ import unicode_literals

from PyQt5.QtCore import QThread
from PyQt5.QtCore import QTimer

from ninja_ide.core import file_manager
from ninja_ide.core import settings
from ninja_ide.dependencies import pycodestylemod


class Pep8Checker(QThread):

    def __init__(self, editor):
        QThread.__init__(self)
        self._editor = editor
        self._path = ''
        self._encoding = ''
        self.pep8checks = {}

    def check_style(self):
        if not self.isRunning():
            self._path = self._editor.ID
            self._encoding = self._editor.encoding
            QTimer.singleShot(1000, self.start)
            # self.start()

    def reset(self):
        self.pep8checks = {}

    def run(self):
        exts = settings.SYNTAX.get('python')['extension']
        file_ext = file_manager.get_file_extension(self._path)
        if file_ext in exts:
            try:
                self.reset()
                source = self._editor.get_text()
                pep8_style = pycodestylemod.StyleGuide(parse_argv=False, config_file='', checker_class=CustomChecker)
                temp_data = pep8_style.input_file(self._path, lines=source.splitlines(True))
                # source_lines = source.split('\n')
                for line, col, code, text in temp_data:
                    message = '[PEP8]: %s' % text
                    self.pep8checks[line - 1] = [message]
                    # print(source_lines[line - 1])
            except Exception as reason:
                print("Checker not finished: {}".format(reason))
            # self.reset()
            # source = self._editor.get_text()
            # # tempData = pycodestylemod.run_check(self._path, source)
            # pep8 = pycodestylemod.StyleGuide(parse_argv=False, config_file='')
            # tempData = pep8.input_file(self._path, lines=source.splitlines(True))
            # i = 0
            # while i < len(tempData):
            #     lineno = -1
            #     try:
            #         offset = 2 + len(file_ext)
            #         startPos = tempData[i].find('.%s:' % file_ext) + offset
            #         endPos = tempData[i].find(':', startPos)
            #         lineno = int(tempData[i][startPos:endPos]) - 1
            #         error = tempData[i][tempData[i].find(
            #             ':', endPos + 1) + 2:]
            #         line = '\n'.join(
            #             [error, tempData[i + 1], tempData[i + 2]])
            #     except Exception:
            #         line = ''
            #     finally:
            #         i += 3
            #     if line and lineno > -1:
            #         if lineno not in self.pep8checks:
            #             self.pep8checks[lineno] = [line]
            #         else:
            #             message = self.pep8checks[lineno]
            #             message += [line]
            #             self.pep8checks[lineno] = message
        else:
            self.reset()


class CustomReport(pycodestylemod.StandardReport):

    def get_file_results(self):
        data = []
        for line_number, offset, code, text, doc in self._deferred_print:
            col = offset + 1
            data.append((line_number, col, code, text))
        return data


class CustomChecker(pycodestylemod.Checker):

    def __init__(self, *args, **kw):
        super().__init__(*args, report=CustomReport(kw.pop("options")), **kw)



# -*- coding: utf-8 -*-

from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QListWidget
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QSpacerItem
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtCore import QSettings
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QSize
# from PyQt5.QtCore import SIGNAL

from ninja_ide import resources
from ninja_ide.core import settings


class PythonDetectDialog(QDialog):

    def __init__(self, suggested, parent=None):
        super(PythonDetectDialog, self).__init__(parent, Qt.Dialog)
        self.setMaximumSize(QSize(0, 0))
        self.setWindowTitle("Configure Python Path")

        vbox = QVBoxLayout(self)

        lblMessage = QLabel(self.tr("We have detected that you are using " +
            "Windows,\nplease choose the proper " +
            "Python application for you:"))
        vbox.addWidget(lblMessage)

        self.listPaths = QListWidget()
        self.listPaths.setSelectionMode(QListWidget.SingleSelection)
        vbox.addWidget(self.listPaths)

        hbox = QHBoxLayout()
        hbox.addSpacerItem(QSpacerItem(1, 0, QSizePolicy.Expanding))
        btnCancel = QPushButton(self.tr("Cancel"))
        btnAccept = QPushButton(self.tr("Accept"))
        hbox.addWidget(btnCancel)
        hbox.addWidget(btnAccept)
        vbox.addLayout(hbox)

        self.connect(btnAccept, SIGNAL("clicked()"), self._set_python_path)
        self.connect(btnCancel, SIGNAL("clicked()"), self.close)

        for path in suggested:
            self.listPaths.addItem(path)
        self.listPaths.setCurrentRow(0)

    def _set_python_path(self):
        python_path = self.listPaths.currentItem().text()

        qsettings = QSettings(resources.SETTINGS_PATH, QSettings.IniFormat)
        settings.PYTHON_PATH = python_path
        settings.PYTHON_PATH_CONFIGURED_BY_USER = True
        qsettings.setValue('preferences/execution/pythonPath', python_path)
        qsettings.setValue('preferences/execution/pythonPathConfigured', True)
        self.close()


# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QListWidget
from PyQt5.QtWidgets import QListWidgetItem
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QSpacerItem
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtWidgets import QPlainTextEdit
from PyQt5.QtGui import QTextCursor
from PyQt5.QtCore import Qt
# from PyQt5.QtCore import SIGNAL

from ninja_ide.gui.main_panel import main_container


class MigrationWidget(QWidget):

    def __init__(self):
        super(MigrationWidget, self).__init__()
        self._migration = {}
        vbox = QVBoxLayout(self)
        lbl_title = QLabel(self.tr("Current code:"))
        self.current_list = QListWidget()
        lbl_suggestion = QLabel(self.tr("Suggested changes:"))
        self.suggestion = QPlainTextEdit()
        self.suggestion.setReadOnly(True)

        self.btn_apply = QPushButton(self.tr("Apply change!"))
        hbox = QHBoxLayout()
        hbox.addSpacerItem(QSpacerItem(1, 0, QSizePolicy.Expanding))
        hbox.addWidget(self.btn_apply)

        vbox.addWidget(lbl_title)
        vbox.addWidget(self.current_list)
        vbox.addWidget(lbl_suggestion)
        vbox.addWidget(self.suggestion)
        vbox.addLayout(hbox)

        self.current_list.itemClicked.connect(self.load_suggestion)
        # self.connect(self.current_list,
        #     SIGNAL("itemClicked(QListWidgetItem*)"), self.load_suggestion)
        self.btn_apply.clicked.connect(self.apply_changes)
        # self.connect(self.btn_apply, SIGNAL("clicked()"), self.apply_changes)

    def apply_changes(self):
        lineno = int(self.current_list.currentItem().data(Qt.UserRole))
        lines = self._migration.migration_data[lineno][0].split('\n')
        remove = -1
        code = ''
        for line in lines:
            if line.startswith('-'):
                remove += 1
            elif line.startswith('+'):
                code += '%s\n' % line[1:]

        editorWidget = main_container.MainContainer().get_actual_editor()
        block_start = editorWidget.document().findBlockByLineNumber(lineno)
        block_end = editorWidget.document().findBlockByLineNumber(
            lineno + remove)
        cursor = editorWidget.textCursor()
        cursor.setPosition(block_start.position())
        cursor.setPosition(block_end.position(), QTextCursor.KeepAnchor)
        cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
        cursor.insertText(code[:-1])

    def load_suggestion(self, item):
        lineno = int(item.data(Qt.UserRole))
        lines = self._migration.migration_data[lineno][0].split('\n')
        code = ''
        for line in lines:
            if line.startswith('+'):
                code += '%s\n' % line[1:]
        self.suggestion.setPlainText(code)
        editorWidget = main_container.MainContainer().get_actual_editor()
        if editorWidget:
            editorWidget.jump_to_line(lineno)
            editorWidget.setFocus()

    def refresh_lists(self, migration):
        self._migration = migration
        self.current_list.clear()
        base_lineno = -1
        for lineno in sorted(migration.migration_data.keys()):
            linenostr = 'L%s\n' % str(lineno + 1)
            data = migration.migration_data[lineno]
            lines = data[0].split('\n')
            if base_lineno == data[1]:
                continue
            base_lineno = data[1]
            message = ''
            for line in lines:
                if line.startswith('-'):
                    message += '%s\n' % line
            item = QListWidgetItem(linenostr + message)
            item.setToolTip(linenostr + message)
            item.setData(Qt.UserRole, lineno)
            self.current_list.addItem(item)

    def clear(self):
        """
        Clear the widget
        """
        self.current_list.clear()
        self.suggestion.clear()



# -*- coding: utf-8 -*-
import os
# from urlparse import urlparse, urlunparse
from urllib.parse import urlparse, urlunparse

from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtCore import QUrl
from PyQt5.QtCore import QDir
from PyQt5.QtCore import QSettings
from PyQt5.QtCore import pyqtSignal
# from PyQt5.QtCore import SIGNAL
# from PyQt5.QtDeclarative import QDeclarativeView
from PyQt5.QtQuickWidgets import QQuickWidget

from ninja_ide import resources
from ninja_ide.gui.main_panel import itab_item


class StartPage(QWidget, itab_item.ITabItem):
    openPreferences = pyqtSignal()
    openProject = pyqtSignal(str)

    def __init__(self, parent=None):
        super(StartPage, self).__init__(parent)
        self._id = "Start Page"
        vbox = QVBoxLayout(self)
        self.view = QQuickWidget()
        self.view.setMinimumWidth(400)
        self.view.setResizeMode(QQuickWidget.SizeRootObjectToView)
        path_qml = QDir.fromNativeSeparators(
            os.path.join(resources.QML_FILES, "StartPage.qml"))
        path_qml = urlunparse(urlparse(path_qml)._replace(scheme='file'))
        self.view.setSource(QUrl(path_qml))
        self.root = self.view.rootObject()
        vbox.addWidget(self.view)

        self.load_items()

        self.root.openProject.connect(self._open_project)
        # self.connect(self.root, SIGNAL("openProject(QString)"),
        #     self._open_project)
        self.root.removeProject.connect(self._on_click_on_delete)
        # self.connect(self.root, SIGNAL("removeProject(QString)"),
        #     self._on_click_on_delete)
        self.root.markAsFavorite.connect(self._on_click_on_favorite)
        # self.connect(self.root, SIGNAL("markAsFavorite(QString, bool)"),
        #     self._on_click_on_favorite)
        self.root.openPreferences.connect(self.openPreferences.emit)

        # self.connect(self.root, SIGNAL("openPreferences()"),
        #     lambda: self.emit(SIGNAL("openPreferences()")))

    def _open_project(self, path):
        # self.emit(SIGNAL("openProject(QString)"), path)
        self.openProject.emit(path)

    def _on_click_on_delete(self, path):
        settings = QSettings(resources.SETTINGS_PATH, QSettings.IniFormat)
        recent_projects = settings.value("recentProjects")
        if path in recent_projects:
            del recent_projects[path]
            settings.setValue("recentProjects", recent_projects)

    def _on_click_on_favorite(self, path, value):
        settings = QSettings(resources.SETTINGS_PATH, QSettings.IniFormat)
        recent_projects = settings.value("recentProjects")
        properties = recent_projects[path]
        properties["isFavorite"] = value
        recent_projects[path] = properties
        settings.setValue("recentProjects", recent_projects)

    def load_items(self):
        settings = QSettings(resources.SETTINGS_PATH, QSettings.IniFormat)
        listByFavorites = []
        listNoneFavorites = []
        recent_projects_dict = dict(settings.value('recentProjects', {}))
        #Filter for favorites
        for recent_project_path, content in list(recent_projects_dict.items()):
            if bool(dict(content)["isFavorite"]):
                listByFavorites.append((recent_project_path,
                    content["lastopen"]))
            else:
                listNoneFavorites.append((recent_project_path,
                    content["lastopen"]))
        if len(listByFavorites) > 1:
            # sort by date favorites
            listByFavorites = sorted(listByFavorites,
                key=lambda date: listByFavorites[1])

        if len(listNoneFavorites) > 1:
            #sort by date last used
            listNoneFavorites = sorted(listNoneFavorites,
                key=lambda date: listNoneFavorites[1])

        for recent_project_path in listByFavorites:
            path = recent_project_path[0]
            name = recent_projects_dict[path]['name']
            self.root.add_project(name, path, True)

        for recent_project_path in listNoneFavorites:
            path = recent_project_path[0]
            name = recent_projects_dict[path]['name']
            self.root.add_project(name, path, False)


# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

from PyQt5.QtGui import QIcon
from PyQt5.QtGui import QKeySequence
from PyQt5.QtGui import QWheelEvent
# from PyQt5.QtCore import SIGNAL
from PyQt5.QtCore import QObject
from PyQt5.QtCore import QPoint
from PyQt5.QtCore import Qt

from ninja_ide import resources


class MenuView(QObject):

    def __init__(self, menuView, toolbar, ide):
        QObject.__init__(self)
        self.__ide = ide

        self.hideConsoleAction = menuView.addAction(
            (self.tr("Show/Hide &Console (%s)") %
                resources.get_shortcut("Hide-misc").toString(
                    QKeySequence.NativeText)))
        self.hideConsoleAction.setCheckable(True)
        self.hideEditorAction = menuView.addAction(
            (self.tr("Show/Hide &Editor (%s)") %
                resources.get_shortcut("Hide-editor").toString(
                    QKeySequence.NativeText)))
        self.hideEditorAction.setCheckable(True)
        self.hideAllAction = menuView.addAction(
            (self.tr("Show/Hide &All (%s)") %
                resources.get_shortcut("Hide-all").toString(
                    QKeySequence.NativeText)))
        self.hideAllAction.setCheckable(True)
        self.hideExplorerAction = menuView.addAction(
            (self.tr("Show/Hide &Explorer (%s)") %
                resources.get_shortcut("Hide-explorer").toString(
                    QKeySequence.NativeText)))
        self.hideExplorerAction.setCheckable(True)
        self.hideToolbarAction = menuView.addAction(
            self.tr("Show/Hide &Toolbar"))
        self.hideToolbarAction.setCheckable(True)
        self.fullscreenAction = menuView.addAction(
            QIcon(resources.IMAGES['fullscreen']),
            (self.tr("Full Screen &Mode (%s)") %
                resources.get_shortcut("Full-screen").toString(
                    QKeySequence.NativeText)))
        menuView.addSeparator()
        splitTabHAction = menuView.addAction(
            QIcon(resources.IMAGES['splitH']),
            (self.tr("Split Tabs Horizontally (%s)") %
                resources.get_shortcut("Split-horizontal").toString(
                    QKeySequence.NativeText)))
        splitTabVAction = menuView.addAction(
            QIcon(resources.IMAGES['splitV']),
            (self.tr("Split Tabs Vertically (%s)") %
                resources.get_shortcut("Split-vertical").toString(
                    QKeySequence.NativeText)))
        followModeAction = menuView.addAction(
            QIcon(resources.IMAGES['follow']),
            (self.tr("Follow Mode (%s)") %
                resources.get_shortcut("Follow-mode").toString(
                    QKeySequence.NativeText)))
        groupTabsAction = menuView.addAction(
            self.tr("Group Tabs by Project"))
        deactivateGroupTabsAction = menuView.addAction(
            self.tr("Deactivate Group Tabs"))
        menuView.addSeparator()
        #Zoom
        zoomInAction = menuView.addAction(QIcon(resources.IMAGES['zoom-in']),
            self.tr("Zoom &In (Ctrl+Wheel-Up)"))
        zoomOutAction = menuView.addAction(QIcon(resources.IMAGES['zoom-out']),
            self.tr("Zoom &Out (Ctrl+Wheel-Down)"))
        menuView.addSeparator()
        fadeInAction = menuView.addAction(
            self.tr("Fade In (Alt+Wheel-Up)"))
        fadeOutAction = menuView.addAction(
            self.tr("Fade Out (Alt+Wheel-Down)"))

        self.toolbar_items = {
            'splitv': splitTabVAction,
            'splith': splitTabHAction,
            'follow-mode': followModeAction,
            'zoom-in': zoomInAction,
            'zoom-out': zoomOutAction
        }

        self.hideConsoleAction.triggered.connect(self.__ide.actions.view_misc_visibility)
        # self.connect(self.hideConsoleAction, SIGNAL("triggered()"),
        #     self.__ide.actions.view_misc_visibility)
        self.hideEditorAction.triggered.connect(self.__ide.actions.view_main_visibility)
        # self.connect(self.hideEditorAction, SIGNAL("triggered()"),
        #     self.__ide.actions.view_main_visibility)
        self.hideExplorerAction.triggered.connect(self.__ide.actions.view_explorer_visibility)
        # self.connect(self.hideExplorerAction, SIGNAL("triggered()"),
        #     self.__ide.actions.view_explorer_visibility)
        self.hideAllAction.triggered.connect(self.__ide.actions.hide_all)
        # self.connect(self.hideAllAction, SIGNAL("triggered()"),
        #     self.__ide.actions.hide_all)
        self.fullscreenAction.triggered.connect(self.__ide.actions.fullscreen_mode)
        # self.connect(self.fullscreenAction, SIGNAL("triggered()"),
        #     self.__ide.actions.fullscreen_mode)

        splitTabHAction.triggered.connect(lambda: self.__ide.mainContainer.split_tab(True))
        # self.connect(splitTabHAction, SIGNAL("triggered()"),
        #     lambda: self.__ide.mainContainer.split_tab(True))
        splitTabVAction.triggered.connect(lambda: self.__ide.mainContainer.split_tab(False))
        # self.connect(splitTabVAction, SIGNAL("triggered()"),
        #     lambda: self.__ide.mainContainer.split_tab(False))
        followModeAction.triggered.connect(self.__ide.mainContainer.show_follow_mode)
        # QObject.connect(followModeAction, SIGNAL("triggered()"),
        #     self.__ide.mainContainer.show_follow_mode)
        zoomInAction.triggered.connect(self.zoom_in_editor)
        # self.connect(zoomInAction, SIGNAL("triggered()"),
        #     self.zoom_in_editor)
        zoomOutAction.triggered.connect(self.zoom_out_editor)
        # self.connect(zoomOutAction, SIGNAL("triggered()"),
        #     self.zoom_out_editor)
        fadeInAction.triggered.connect(self._fade_in)
        # self.connect(fadeInAction, SIGNAL("triggered()"), self._fade_in)
        fadeOutAction.triggered.connect(self._fade_out)
        # self.connect(fadeOutAction, SIGNAL("triggered()"), self._fade_out)
        self.hideToolbarAction.triggered.connect(self._hide_show_toolbar)
        # self.connect(self.hideToolbarAction, SIGNAL("triggered()"),
        #     self._hide_show_toolbar)
        groupTabsAction.triggered.connect(self.__ide.actions.group_tabs_together)
        # self.connect(groupTabsAction, SIGNAL("triggered()"),
        #     self.__ide.actions.group_tabs_together)
        deactivateGroupTabsAction.triggered.connect(self.__ide.actions.deactivate_tabs_groups)
        # self.connect(deactivateGroupTabsAction, SIGNAL("triggered()"),
        #     self.__ide.actions.deactivate_tabs_groups)

        #Set proper MenuView checks values:
        self.hideAllAction.setChecked(True)
        self.hideConsoleAction.setChecked(False)
        self.hideEditorAction.setChecked(True)
        self.hideExplorerAction.setChecked(True)
        self.hideToolbarAction.setChecked(True)

    def _hide_show_toolbar(self):
        if self.__ide.toolbar.isVisible():
            self.__ide.toolbar.hide()
        else:
            self.__ide.toolbar.show()

    def zoom_in_editor(self):
        editor = self.__ide.mainContainer.get_actual_editor()
        if editor:
            editor.zoom_in()

    def zoom_out_editor(self):
        editor = self.__ide.mainContainer.get_actual_editor()
        if editor:
            editor.zoom_out()

    def _fade_in(self):
        event = QWheelEvent(QPoint(), 120, Qt.NoButton, Qt.AltModifier)
        self.__ide.wheelEvent(event)

    def _fade_out(self):
        event = QWheelEvent(QPoint(), -120, Qt.NoButton, Qt.AltModifier)
        self.__ide.wheelEvent(event)



#!/usr/bin/env python3
# -*-coding:utf-8-*-

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


###############################################################################
# DOCS
###############################################################################

"""Setup for Ninja-ide (http://www.ninja-ide.org)

NINJA-IDE is a cross-platform integrated development environment (IDE).
NINJA-IDE runs on Linux/X11, Mac OS X and Windows desktop operating systems,
and allows developers to create applications for several purposes using all the
tools and utilities of NINJA-IDE, making the task of writing software easier
and more enjoyable.
"""


###############################################################################
# IMPORTS
###############################################################################

# import sys
import os
# import shutil

from distutils.command.install import install
from setuptools import setup, find_packages

import ninja_ide


class CustomInstall(install):
    """
    Custom installation class on package files.

    It copies all the files into the "PREFIX/share/ninja-ide" dir.
    """

    def run(self):
        install.run(self)

        for script in self.distribution.scripts:

            script_path = os.path.join(self.install_scripts,
                                       os.path.basename(script))

            with open(script_path, 'r') as f:
                content = f.read()
            content = content.replace('@ INSTALLED_BASE_DIR @',
                                      self._custom_data_dir)
            with open(script_path, 'w') as f:
                f.write(content)

            src_desktop = self.distribution.get_name() + '.desktop'
            src_desktop = src_desktop.lower()

            if not os.path.exists(self._custom_apps_dir):
                os.makedirs(self._custom_apps_dir)
            dst_desktop = os.path.join(self._custom_apps_dir, src_desktop)
            with open(src_desktop, 'r') as f:
                content = f.read()
            icon = os.path.join(self._custom_data_dir, "ninja_ide", "img", "icon.png")
            content = content.replace('@ INSTALLED_ICON @', icon)
            with open(dst_desktop, 'w') as f:
                f.write(content)

            # Man dir
            # if not os.path.exists(self._custom_man_dir):
            #     os.makedirs(self._custom_man_dir)
            # shutil.copy("man/ninja-ide.1", self._custom_man_dir)

    def finalize_options(self):
        """ Alter the installation path """

        install.finalize_options(self)

        data_dir = os.path.join(self.prefix, "share",
                                self.distribution.get_name())
        apps_dir = os.path.join(self.prefix, "share", "applications")
        # man_dir = os.path.join(self.prefix, "share", "man", "man1")

        if self.root is None:
            build_dir = data_dir
        else:

            build_dir = os.path.join(self.root, data_dir[1:])
            apps_dir = os.path.join(self.root, apps_dir[1:])
            # man_dir = os.path.join(self.root, man_dir[1:])

        self.install_lib = build_dir

        self._custom_data_dir = data_dir
        self._custom_apps_dir = apps_dir
        # self._custom_man_dir = man_dir


###############################################################################
# SETUP
###############################################################################

setup(
    name=ninja_ide.__prj__,
    version=ninja_ide.__version__,
    description=ninja_ide.__doc__,
    author=ninja_ide.__author__,
    author_email=ninja_ide.__mail__,
    url=ninja_ide.__url__,
    license=ninja_ide.__license__,
    keywords="ide python ninja development",
    classifiers=[
        "Development Status :: Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3"
    ],
    package_data={
        "ninja_ide": [
            "addins/lang/*",
            "addins/qml/*",
            "addins/qml/img/*",
            "addins/syntax/*",
            "addins/theme/*",
            "img/*"
        ]
    },
    packages=find_packages(exclude=["ninja_tests", "debian"]),
    scripts=["ninja-ide"],
    cmdclass={"install": CustomInstall}
)


###############################################################################
# MAIN
###############################################################################

if __name__ == '__main__':
    print(__doc__)



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QTreeWidget
from PyQt5.QtWidgets import QTreeWidgetItem
from PyQt5.QtWidgets import QAbstractItemView
from PyQt5.QtWidgets import QHeaderView
from PyQt5.QtCore import Qt
# from PyQt5.QtCore import SIGNAL

from ninja_ide.gui.main_panel import main_container


class Results(QWidget):

    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self._parent = parent
        vbox = QVBoxLayout(self)
        self._tree = QTreeWidget()
        self._tree.setHeaderLabels((self.tr("Content"),
            self.tr('File'), self.tr('Line')))
        self._tree.header().setHorizontalScrollMode(
            QAbstractItemView.ScrollPerPixel)
        self._tree.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self._tree.header().setSectionResizeMode(1, QHeaderView.ResizeToContents)
        self._tree.header().setSectionResizeMode(2, QHeaderView.ResizeToContents)
        self._tree.header().setStretchLastSection(True)
        self._tree.sortByColumn(1, Qt.AscendingOrder)

        vbox.addWidget(self._tree)

        #Signals
        self._tree.itemActivated.connect(self._open_result)
        # self.connect(self._tree,
        #     SIGNAL("itemActivated(QTreeWidgetItem*, int)"),
        #     self._open_result)
        self._tree.itemClicked.connect(self._open_result)
        # self.connect(self._tree, SIGNAL("itemClicked(QTreeWidgetItem*, int)"),
        #     self._open_result)

    def _open_result(self, item, col):
        filename = item.toolTip(1)
        line = int(item.text(2)) - 1
        main_container.MainContainer().open_file(
                filename=filename,
                cursorPosition=line,
                positionIsLineNumber=True)
        self._parent.hide()

    def update_result(self, items):
        self._tree.clear()
        for i in items:
            item = QTreeWidgetItem(self._tree, (i[3], i[0], str(i[2] + 1)))
            item.setToolTip(1, i[1])



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QTabWidget
from PyQt5.QtWidgets import QPlainTextEdit
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QPushButton
# from PyQt5.QtCore import SIGNAL


class PluginErrorDialog(QDialog):
    """
    Dialog with tabs each tab is a python traceback
    """
    def __init__(self):
        QDialog.__init__(self)
        self.setWindowTitle(self.tr("Plugin error report"))
        self.resize(525, 400)
        vbox = QVBoxLayout(self)
        label = QLabel(self.tr('Some plugins have errors and were removed'))
        vbox.addWidget(label)
        self._tabs = QTabWidget()
        vbox.addWidget(self._tabs)
        hbox = QHBoxLayout()
        btnAccept = QPushButton(self.tr("Accept"))
        btnAccept.setMaximumWidth(100)
        hbox.addWidget(btnAccept)
        vbox.addLayout(hbox)
        #signals
        self.connect(btnAccept, SIGNAL("clicked()"), self.close)

    def add_traceback(self, plugin_name, traceback_msg):
        traceback_widget = TracebackWidget(traceback_msg)
        self._tabs.addTab(traceback_widget, plugin_name)


class TracebackWidget(QWidget):
    """
    Represents a python traceback
    """

    def __init__(self, traceback_msg):
        QWidget.__init__(self)
        vbox = QVBoxLayout(self)
        self._editor = QPlainTextEdit()
        vbox.addWidget(QLabel(self.tr('Traceback')))
        vbox.addWidget(self._editor)
        self._editor.setReadOnly(True)
        self._editor.insertPlainText(traceback_msg)



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
# Based on https://github.com/DSpeckhals/python-indent

# from ninja_ide.gui.editor.base_inde import base
from PyQt5.QtGui import QTextCursor
from ninja_ide.core import settings
from ninja_ide.tools.logger import NinjaLogger
# Logger
logger = NinjaLogger(__name__)


class BaseIndenter(object):

    def __init__(self, neditor):
        self._neditor = neditor  # Editor ref
        self.width = settings.INDENT
        self.use_tabs = settings.USE_TABS

    def text(self):
        """Get indent text as \t or string of spaces"""
        text = " " * self.width
        if self.use_tabs:
            text = "\t"
        return text

    def indent_block(self, cursor):
        """Indent block after enter pressed"""

        at_start_of_line = cursor.positionInBlock() == 0
        with self._neditor:
            cursor.insertBlock()
            if not at_start_of_line:
                indent = self._compute_indent(cursor)
                if indent is not None:
                    cursor.insertText(indent)
                    return True
                return False
        self._neditor.ensureCursorVisible()

    def block_indent(self, block):
        block_text = block.text()
        space_at_start_len = len(block_text) - len(block_text.lstrip())
        return block_text[:space_at_start_len]

    def indent(self):
        """Indent a single line after tab pressed"""
        cursor = self._neditor.textCursor()
        if self.use_tabs:
            to_insert = self.text()
        else:
            text = cursor.block().text()
            text_before_cursor = text[:cursor.positionInBlock()]
            to_insert = self.width - (len(text_before_cursor)) % self.width
            to_insert *= " "
        cursor.insertText(to_insert)

    def indent_selection(self):
        """Indent selection after tab pressed"""

        def indent_block(block):
            cursor = QTextCursor(block)
            indentation = self.block_indent(block)
            cursor.setPosition(block.position() + len(indentation))
            cursor.insertText(self.text())

        cursor = self._neditor.textCursor()
        start_block = self._neditor.document().findBlock(
            cursor.selectionStart())
        end_block = self._neditor.document().findBlock(
            cursor.selectionEnd())

        with self._neditor:
            if start_block != end_block:
                stop_block = end_block.next()
                # Indent multiple lines
                block = start_block
                while block != stop_block:
                    indent_block(block)
                    block = block.next()
                new_cursor = QTextCursor(start_block)
                new_cursor.setPosition(
                    end_block.position() + len(end_block.text()),
                    QTextCursor.KeepAnchor)
                self._neditor.setTextCursor(new_cursor)
            else:
                # Indent one line
                indent_block(start_block)

    def line_indent(self, line):
        """Returns the indentation level of `line`"""

        if line == -1:
            line = self._neditor.textCursor().blockNumber() + 1
        text = self._neditor.document().findBlockByNumber(line).text()
        indentation = len(text) - len(text.lstrip())
        return indentation

    def _compute_indent(self, cursor):
        raise NotImplementedError


class BasicIndenter(BaseIndenter):
    """Basic indenter"""

    def _compute_indent(self, cursor):
        # At this point, enter was pressed, so the block should be previous
        return self.block_indent(cursor.block().previous())


class PythonIndenter(BaseIndenter):
    """PEP8 indenter for Python"""
    LANG = 'python'

    def _compute_indent(self, cursor):
        # At this point, the new block has added
        block = cursor.block()
        line = self._neditor.textCursor().blockNumber()
        # Source code at current position
        text = self._neditor.get_text()[:cursor.position()]
        current_indent = self.block_indent(block.previous())

        # Parse text
        results = self.__parse_text(text)
        # Get result of parsed text
        bracket_stack, last_closed_line, should_hang, last_colon_line = results
        logger.debug(results)

        if should_hang:
            cursor = self._neditor.textCursor()
            text = cursor.block().text()
            next_char = ""
            if (len(text) > 0):
                next_char = text[0]

            if len(next_char) > 0 and next_char in "}])":
                cursor.insertBlock()
                cursor.insertText(current_indent)
                cursor.movePosition(cursor.Up)
                self._neditor.setTextCursor(cursor)
            else:
                cursor.insertText(current_indent)

            cursor.insertText(current_indent + self.text())
            return

        if not bracket_stack:
            if last_closed_line:
                if last_closed_line[1] == line - 1:
                    indent_level = self.line_indent(last_closed_line[0])
                    if last_colon_line == line - 1:
                        indent_level += self.width
                    return indent_level * " "
                else:
                    text_stripped = block.previous().text().strip()
                    if text_stripped in ("break", "continue", "raise", "pass",
                                         "return") or \
                            text_stripped.startswith("raise ") or \
                            text_stripped.startswith("return "):
                        return " " * (len(current_indent) - self.width)
            else:
                text_stripped = block.previous().text().strip()
                if text_stripped in ("break", "continue", "raise", "pass",
                                     "return") or \
                        text_stripped.startswith("raise ") or \
                        text_stripped.startswith("return "):
                    return " " * (len(current_indent) - self.width)

            last_previous_char = block.previous().text()[-1]
            if last_previous_char == ":":
                return self.text() + current_indent
            return self.block_indent(block.previous())

        last_open_bracket = bracket_stack.pop()

        have_closed_bracket = len(last_closed_line) > 0
        just_opened_bracket = last_open_bracket[0] == line - 1
        just_closed_bracket = have_closed_bracket and \
            last_closed_line[1] == line - 1

        v = have_closed_bracket and last_closed_line[0] > last_open_bracket[0]
        if not just_opened_bracket and not just_closed_bracket:
            return current_indent
        elif just_closed_bracket and v:
            previous_indent = self.line_indent(last_closed_line[0])
            indent_col = previous_indent
        else:
            indent_col = last_open_bracket[1] + 1
        return indent_col * " "

    def __parse_text(self, text):
        # [line, column] pairs describing where open brackets are
        open_bracket = []
        # Describing the lines where the last bracket to be closed was
        # opened and closed
        last_close_line = []
        # The last line a def/for/if/elif/else/try/except block started
        last_colon_line = None
        # Indicating wheter or not a hanging indent is needed
        should_hang = False

        for lineno, line in enumerate(text.splitlines()):
            last_last_colon_line = last_colon_line
            for col, char in enumerate(line):
                if char in '{[(':
                    open_bracket.append([lineno, col])
                    should_hang = True
                else:
                    should_hang = False
                    last_colon_line = last_last_colon_line
                    if char == ':':
                        last_colon_line = lineno
                    elif char in '}])' and open_bracket:
                        opened_row = open_bracket.pop()[0]
                        if lineno != opened_row:
                            last_close_line = [opened_row, lineno]
        return (open_bracket, last_close_line, should_hang, last_colon_line)



# -*- coding: utf-8 -*-
#
# This file is part of NINJA-IDE (http://ninja-ide.org).
#
# NINJA-IDE is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# any later version.
#
# NINJA-IDE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NINJA-IDE; If not, see <http://www.gnu.org/licenses/>.
import _ast
import ast

from ninja_ide.tools.completion import analyzer

from ninja_ide.tools.logger import NinjaLogger

logger_imports = NinjaLogger(
    'ninja_ide.tools.introspection.obtaining_imports')
logger_symbols = NinjaLogger(
    'ninja_ide.tools.introspection.obtaining_symbols')

_map_type = {
    _ast.Tuple: 'tuple',
    _ast.List: 'list',
    _ast.Str: 'str',
    _ast.Dict: 'dict',
    _ast.Num: 'int',
    _ast.Call: 'function()',
}


def _parse_assign(symbol):
    assigns = {}
    attributes = {}
    for var in symbol.targets:
        if isinstance(var, ast.Attribute):
            attributes[var.attr] = var.lineno
        elif isinstance(var, ast.Name):
            assigns[var.id] = var.lineno
    return (assigns, attributes)


def _parse_class(symbol, with_docstrings):
    docstring = {}
    attr = {}
    func = {}
    clazz = {}
    name = symbol.name + '('
    name += ', '.join([
        analyzer.expand_attribute(base) for base in symbol.bases])
    name += ')'
    for sym in symbol.body:
        if isinstance(sym, ast.Assign):
            result = _parse_assign(sym)
            attr.update(result[0])
            attr.update(result[1])
        elif isinstance(sym, ast.FunctionDef):
            result = _parse_function(sym, with_docstrings)
            attr.update(result['attrs'])
            if with_docstrings:
                docstring.update(result['docstring'])
            func[result['name']] = {
                'lineno': result['lineno'],
                'functions': result['functions']
            }
        elif isinstance(sym, ast.ClassDef):
            result = _parse_class(sym, with_docstrings)
            clazz[result['name']] = {
                'lineno': result['lineno'],
                'members': {
                    'attributes': result['attributes'],
                    'functions': result['functions']
                    }
                }
            docstring.update(result['docstring'])
    if with_docstrings:
        docstring[symbol.lineno] = ast.get_docstring(symbol, clean=True)

    lineno = symbol.lineno
    for decorator in symbol.decorator_list:
        lineno += 1

    return {
        'name': name,
        'attributes': attr,
        'functions': func,
        'lineno': lineno,
        'docstring': docstring,
        'classes': clazz
    }


def _parse_function(symbol, with_docstrings):
    docstring = {}
    attrs = {}
    func = {'functions': {}}

    func_name = symbol.name + '('
    # We store the arguments to compare with default backwards
    defaults = []
    for value in symbol.args.defaults:
        # TODO: In some cases we can have something like: a=os.path
        defaults.append(value)
    arguments = []
    for arg in reversed(symbol.args.args):
        if not isinstance(arg, _ast.Name) or arg.id == "self":
            continue
        argument = arg.id
        if defaults:
            value = defaults.pop()
            arg_default = _map_type.get(value.__class__, None)
            if arg_default is None:
                if isinstance(value, _ast.Attribute):
                    arg_default = analyzer.expand_attribute(value)
                elif isinstance(value, _ast.Name):
                    arg_default = value.id
                else:
                    arg_default = 'object'
            argument += '=' + arg_default
        arguments.append(argument)
    func_name += ', '.join(reversed(arguments))

    if symbol.args.vararg is not None:
        if not func_name.endswith('('):
            func_name += ', '
        func_name += '*'
        func_name += symbol.args.vararg.arg
    if symbol.args.kwarg is not None:
        if not func_name.endswith('('):
            func_name += ', '
        func_name += '**'
        func_name += symbol.args.kwarg.arg
    func_name += ')'

    for sym in symbol.body:
        if isinstance(sym, ast.Assign):
            result = _parse_assign(sym)
            attrs.update(result[1])
        elif isinstance(sym, ast.FunctionDef):
            result = _parse_function(sym, with_docstrings)
            if with_docstrings:
                docstring.update(result['docstring'])
            func['functions'][result['name']] = {
                'lineno': result['lineno'],
                'functions': result['functions']
            }

    if with_docstrings:
        docstring[symbol.lineno] = ast.get_docstring(symbol, clean=True)

    lineno = symbol.lineno
    for decorator in symbol.decorator_list:
        lineno += 1

    return {'name': func_name, 'lineno': lineno,
            'attrs': attrs, 'docstring': docstring, 'functions': func}


def obtain_symbols(source, with_docstrings=False, filename='',
                   simple=False, only_simple=False):
    """Parse a module source code to obtain: Classes, Functions and Assigns."""

    try:
        module = ast.parse(source)
    except SyntaxError:
        logger_symbols.debug("The file contains syntax errors: %s" % filename)
        if simple:
            return {}, {}
        else:
            return {}
    symbols = {}
    symbols_simplified = {}
    globalAttributes = {}
    globalFunctions = {}
    classes = {}
    docstrings = {}

    for symbol in module.body:
        if isinstance(symbol, ast.Assign) and not only_simple:
            result = _parse_assign(symbol)
            globalAttributes.update(result[0])
            globalAttributes.update(result[1])
        elif isinstance(symbol, ast.FunctionDef):
            if not only_simple:
                result = _parse_function(symbol, with_docstrings)
                if with_docstrings:
                    docstrings.update(result['docstring'])
                globalFunctions[result['name']] = {
                    'lineno': result['lineno'],
                    'functions': result['functions']}
            if simple:
                result_simple = _parse_function_simplified(symbol)
                symbols_simplified.update(result_simple)
        elif isinstance(symbol, ast.ClassDef):
            if not only_simple:
                result = _parse_class(symbol, with_docstrings)
                classes[result['name']] = {
                    'lineno': result['lineno'],
                    'members': {'attributes': result['attributes'],
                                'functions': result['functions'],
                                'classes': result['classes']}}
                docstrings.update(result['docstring'])
            if simple:
                result_simple = _parse_class_simplified(symbol)
                symbols_simplified.update(result_simple)
    if globalAttributes:
        symbols['attributes'] = globalAttributes
    if globalFunctions:
        symbols['functions'] = globalFunctions
    if classes:
        symbols['classes'] = classes
    if docstrings and with_docstrings:
        symbols['docstrings'] = docstrings

    if simple:
        return symbols, symbols_simplified
    else:
        return symbols


def obtain_imports(source='', body=None):
    if source:
        try:
            module = ast.parse(source)
            body = module.body
        except SyntaxError:
            logger_imports.debug("A file contains syntax errors.")
    # Imports{} = {name: asname}, for example = {sys: sysAlias}
    imports = {}
    # From Imports{} = {name: {module: fromPart, asname: nameAlias}}
    fromImports = {}
    if body is not None:
        for sym in body:
            if isinstance(sym, ast.Import):
                for item in sym.names:
                    imports[item.name] = {
                        'asname': item.asname,
                        'lineno': sym.lineno
                    }
            if isinstance(sym, ast.ImportFrom):
                for item in sym.names:
                    fromImports[item.name] = {
                        'module': sym.module,
                        'asname': item.asname,
                        'lineno': sym.lineno
                    }
    return {'imports': imports, 'fromImports': fromImports}


def _parse_class_simplified(symbol):
    results = {}
    name = symbol.name + '('
    name += ', '.join([
        analyzer.expand_attribute(base) for base in symbol.bases])
    name += ')'
    for sym in symbol.body:
        if isinstance(sym, ast.FunctionDef):
            result = _parse_function_simplified(sym, symbol.name)
            results.update(result)
        elif isinstance(sym, ast.ClassDef):
            result = _parse_class_simplified(sym)
            results.update(result)

    lineno = symbol.lineno
    for decorator in symbol.decorator_list:
        lineno += 1

    results[lineno] = (name, 'c')
    return results


def _parse_function_simplified(symbol, member_of=""):
    results = {}
    inside_class = True if member_of != "" else False

    if member_of:
        func_name = member_of + " : " + symbol.name + '('
    else:
        func_name = symbol.name + '('
    # We store the arguments to compare with default backwards
    defaults = []
    for value in symbol.args.defaults:
        # TODO: In some cases we can have something like: a=os.path
        defaults.append(value)
    arguments = []
    for arg in reversed(symbol.args.args):
        if not isinstance(arg, _ast.Name) or arg.id == "self":
            continue
        argument = arg.id
        if defaults:
            value = defaults.pop()
            arg_default = _map_type.get(value.__class__, None)
            if arg_default is None:
                if isinstance(value, _ast.Attribute):
                    arg_default = analyzer.expand_attribute(value)
                elif isinstance(value, _ast.Name):
                    arg_default = value.id
                else:
                    arg_default = 'object'
            argument += '=' + arg_default
        arguments.append(argument)
    func_name += ', '.join(reversed(arguments))
    if symbol.args.vararg is not None:
        if not func_name.endswith('('):
            func_name += ', '
        func_name += '*'
        func_name += symbol.args.vararg.arg
    if symbol.args.kwarg is not None:
        if not func_name.endswith('('):
            func_name += ', '
        func_name += '**'
        func_name += symbol.args.kwarg.arg
    func_name += ')'

    for sym in symbol.body:
        if isinstance(sym, ast.FunctionDef):
            result = _parse_function_simplified(sym, symbol.name)
            results.update(result)

    lineno = symbol.lineno
    for decorator in symbol.decorator_list:
        lineno += 1

    results[lineno] = (func_name, 'f', inside_class)
    return results
